import os
import uvicore
from uvicore.support import str
from uvicore.support.dumper import dump, dd
from uvicore.support.schematic import Schematic
from uvicore.console import command, argument, option


@command()
@argument('name')
@argument('table')
async def model(name: str, table: str):
    """Generate a new ORM model schematic"""

    stub = os.path.dirname(__file__) + '/stubs/model.py'
    dest = uvicore.config('app.paths.models') + '/' + name + '.py'

    Schematic(
        type='model',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_modelname', name),
            ('xx_ModelName', str.studly(name)),
            ('xx_tablename', table),
            ('xx_TableName', str.studly(table)),
        ]
    ).generate()



#from sqlalchemy.ext.declarative import declarative_base
import functools
import sqlalchemy
from sqlalchemy import event, orm
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from typing import Any
# from typing import Iterator, Any
# from functools import lru_cache
# from fastapi_utils.session import FastAPISessionMaker
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.ext.declarative import as_declarative, declared_attr
# from sqlalchemy.orm import Session

# #Model = declarative_base()
# #Model = sa.ext.declarative.declarative_base()

# # def get_db() -> Iterator[Session]:
# #     """ FastAPI dependency that provides a sqlalchemy session """
# #     yield from _sessionmaker().get_db()


# # @lru_cache()
# # def _sessionmaker() -> FastAPISessionMaker:
# #     """ This function could be replaced with a global variable if preferred """
# #     #database_uri = DBSettings().database_uri
# #     database_uri = "mysql://root:techie@127.0.0.1/uvicore_wiki"
# #     return FastAPISessionMaker(database_uri)


# from sqlalchemy import create_engine
# from sqlalchemy.orm import sessionmaker

# engine = create_engine("mysql+pymysql://root:techie@127.0.0.1/uvicore_wiki")
# SessionLocal = sessionmaker(bind=engine)

# @as_declarative()
class Model2:
    id: Any
    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()


class Model(object):
    """Base class for SQLAlchemy declarative base model.

    To define models, subclass :attr:`db.Model <SQLAlchemy.Model>`, not this
    class. To customize ``db.Model``, subclass this and pass it as
    ``model_class`` to :class:`SQLAlchemy`.
    """

    #: Query class used by :attr:`query`. Defaults to
    # :class:`SQLAlchemy.Query`, which defaults to :class:`BaseQuery`.
    query_class = None

    #: Convenience property to query the database for instances of this model
    # using the current session. Equivalent to ``db.session.query(Model)``
    # unless :attr:`query_class` has been changed.
    query = None

    def __repr__(self):
        identity = inspect(self).identity
        if identity is None:
            pk = "(transient {0})".format(id(self))
        else:
            pk = ', '.join(to_str(value) for value in identity)
        return '<{0} {1}>'.format(type(self).__name__, pk)

class BaseQuery(orm.Query):

    def xx(self):
        return 'xx'


def _make_table(db):
    def _make_table(*args, **kwargs):
        if len(args) > 1 and isinstance(args[1], db.Column):
            args = (args[0], db.metadata) + args[1:]
        info = kwargs.pop('info', None) or {}
        info.setdefault('bind_key', None)
        kwargs['info'] = info
        return sqlalchemy.Table(*args, **kwargs)
    return _make_table

def _wrap_with_default_query_class(fn, cls):
    @functools.wraps(fn)
    def newfn(*args, **kwargs):
        _set_default_query_class(kwargs, cls)
        if "backref" in kwargs:
            backref = kwargs['backref']
            if isinstance(backref, string_types):
                backref = (backref, {})
            _set_default_query_class(backref[1], cls)
        return fn(*args, **kwargs)
    return newfn

class _QueryProperty(object):
    def __init__(self, sa):
        self.sa = sa

    def __get__(self, obj, type):
        try:
            mapper = orm.class_mapper(type)
            if mapper:
                return type.query_class(mapper, session=self.sa.session())
        except UnmappedClassError:
            return None

class SQLAlchemy(object):

    def __init__(self, query_class=BaseQuery, model_class=Model2, metadata=None):
        self.engine = self.__create_engine()
        self.Query = query_class
        self.session = orm.sessionmaker(bind=self.engine)
        self.Model = self.make_declarative_base(model_class, metadata)
        self.__include_sqlalchemy(query_class)

    def make_declarative_base(self, model, metadata=None):
        if not isinstance(model, DeclarativeMeta):
            model = declarative_base(
                cls=model,
                name='Model',
                metadata=metadata,
                #metaclass=DefaultMeta
            )
        if metadata is not None and model.metadata is not metadata:
            model.metadata = metadata

        if not getattr(model, 'query_class', None):
            model.query_class = self.Query

        model.query = _QueryProperty(self)
        return model

    def __include_sqlalchemy(self, cls):
        for module in sqlalchemy, sqlalchemy.orm:
            for key in module.__all__:
                if not hasattr(self, key):
                    setattr(self, key, getattr(module, key))
        # Note: self.Table does not attempt to be a SQLAlchemy Table class.
        self.Table = _make_table(self)
        self.relationship = _wrap_with_default_query_class(self.relationship, cls)
        self.relation = _wrap_with_default_query_class(self.relation, cls)
        self.dynamic_loader = _wrap_with_default_query_class(self.dynamic_loader, cls)
        self.event = event

    def create_all(self):
        self.Model.metedata.create_all(bind=self.engine)

    def __create_engine(self):
        return sqlalchemy.create_engine("mysql+pymysql://root:techie@127.0.0.1/uvicore_wiki")



from __future__ import annotations
import jinja2
import uvicore
from typing import Dict, List, Any
from starlette.templating import Jinja2Templates as _Jinja2Templates
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Template as TemplateInterface


@uvicore.service('uvicore.http.templating.jinja.Jinja',
    aliases=['Templates', 'templates'],
    singleton=True,
)
class Jinja(TemplateInterface, _Jinja2Templates):
    """Jinja Template private class.

    Dont access templates directly.
    Use reponse.View() instead helper instead."""

    @property
    def env(self) -> jinja2.Environment:
        return self._env

    @property
    def paths(self) -> List[str]:
        return self._paths

    @property
    def context_functions(self) -> Dict:
        return self._context_functions

    @property
    def context_filters(self) -> Dict:
        return self._context_filters

    @property
    def filters(self) -> Dict:
        return self._filters

    @property
    def tests(self) -> Dict:
        return self._tests

    def __init__(self) -> None:
        self._env = None
        self._paths = []
        self._context_functions = {}
        self._context_filters = {}
        self._filters = {}
        self._tests = {}

    def init(self) -> None:
        # Load our jinja2 environment
        loader = jinja2.FileSystemLoader(self.paths)
        self._env = jinja2.Environment(loader=loader, autoescape=True)

        # Add user defined options
        self._register_options()

    def _register_options(self):
        # Add Context Functions
        for name, method in self.context_functions.items():
            self._env.globals[name] = jinja2.contextfunction(method)

        # Add Context Filters
        for name, method in self.context_filters.items():
            self._env.filters[name] = jinja2.contextfilter(method)

        # Add Filters
        for name, method in self.filters.items():
            self._env.filters[name] = method

        # Add Tests
        for name, method in self.tests.items():
            self._env.tests[name] = method

    def include_path(self, path) -> None:
        if path not in self.paths:
            self._paths.append(path)

    def include_context_function(self, name: str, method: Any) -> None:
        self._context_functions[name] = method

    def include_context_filter(self, name: str, method: Any) -> None:
        self._context_filters[name] = method

    def include_filter(self, name: str, method: Any) -> None:
        self._filters[name] = method

    def include_test(self, name: str, method: Any) -> None:
        self._tests[name] = method

# IoC Class Instance
# No, not to be used by public by importing.  Use ioc.make instead.


# Public API for import * and doc gens
#__all__ = ['_Jinja']



import uvicore
import aiohttp
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd


@uvicore.provider()
class HttpClient(ServiceProvider):

    def register(self) -> None:

        # Register event listeners
        # String based events instead of class based because HTTP may not even
        # be installed, so importing it would cause an issue.
        @uvicore.events.handle(['uvicore.console.events.command.Startup', 'uvicore.http.events.server.Startup'])
        async def uvicore_startup(event):
            #print('aiohttp console/http startup')
            # Fire up an aiohttp client session and bind to IoC as a singleton
            uvicore.ioc.bind('uvicore.http_client', aiohttp.ClientSession(),
                aliases=['http_client', 'aiohttp'],
                singleton=True
            )

        @uvicore.events.handle(['uvicore.console.events.command.Shutdown', 'uvicore.http.events.server.Shutdown'])
        async def uvicore_shutdown(event):
            #print('aiohttp console/http shutdown')
            # Shutdown the aiohttp client session
            await uvicore.ioc.make('aiohttp').close()

    def boot(self) -> None:
        pass



import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd

from uvicore.database.provider import Db

#JUNK
from uvicore.http.provider import Http

@uvicore.provider()
class Auth(ServiceProvider, Db, Http):

    def register(self) -> None:
        # Register IoC bindings
        # self.bind(
        #     name='Auth',
        #     object='uvicore.auth.auth._Auth',
        #     aliases=['auth']
        # )

        # Bind Tables
        #self.bind('uvicore.auth.database.tables.groups.Groups', 'uvicore.auth.database.tables.groups._Groups', singleton=True)
        #self.bind('uvicore.auth.database.tables.user_info.UserInfo', 'uvicore.auth.database.tables.user_info._UserInfo', singleton=True)
        #self.bind('uvicore.auth.database.tables.users.Users', 'uvicore.auth.database.tables.users._Users', singleton=True)

        # Bind Models
        #self.bind('uvicore.auth.models.group.Group', 'uvicore.auth.models.group.GroupModel')
        #self.bind('uvicore.auth.models.user.User', 'uvicore.auth.models.user.UserModel')
        #self.bind('uvicore.auth.models.user_info.UserInfo', 'uvicore.auth.models.user_info.UserInfoModel')

        # Register config
        self.configs([
            {'key': self.name, 'module': 'uvicore.auth.config.package.config'}
        ])

    def boot(self) -> None:
        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define Database Connections
        self.connections(
            connections=self.package.config.database.connections,
            default=self.package.config.database.default
        )

        # Define all tables/models used by this package.
        # The goal is to load up all SQLAlchemy tables for complete metedata definitions.
        # If you separate tables vs models use self.tables(['myapp.database.tables.*])
        # If you use models only, or models with inline tables then use self.models(['myapp.models.*])
        # Optionally define your own models/__init__.py and import myapp.models to load every table.
        # Order does not matter as they are sorted topologically for ForeignKey dependencies
        # Using __init__.py now, so just import it
        #from uvicore.auth import models
        #dump(self.package)
        self.models(['uvicore.auth.models'])

        # Define data seeders
        # NO - Auth shouldn't do its own seeding.  Let the app do it all.
        # You think?  What if a package is an app, then it runs seeders, but if that app is used
        # inside another package, you can't stop it from seeding.  Need to figure out overrideing seeders array better
        self.seeders(['uvicore.auth.database.seeders.seed'])

        # Define view and asset paths and configure the templating system
        self.define_views()

        # Define Web and API routes and prefixes
        self.define_routes()

    def define_views(self) -> None:
        """Define view and asset paths and configure the templating system"""

        # Define view paths
        #self.views(['uvicore.auth.http.views'])

        # Define public paths
        self.public(['uvicore.auth.http.public'])

        # Define asset paths
        self.assets(['uvicore.auth.http.public.assets'])

    def define_routes(self) -> None:
        """Define Web and API routes and prefixes"""

        # Define web routes
        # self.web_routes(
        #     module='uvicore.auth.http.routes.web.Web',
        #     prefix=self.package.config.web.prefix,
        #     #name_prefix=None,
        # )

        # Define api routes
        self.api_routes(
            module='uvicore.auth.http.routes.api.Api',
            prefix=self.package.config.api.prefix,
            #name_prefix='api',
        )



import uvicore
from uvicore.typing import List, Union
from uvicore.contracts import User as UserInterface
#from uvicore.contracts import UserProvider
from uvicore.support.dumper import dump, dd
#from uvicore.support.hash import sha1
#from uvicore.auth.support import password as pwd


@uvicore.service()
class User(UserInterface):
    """Auth logged in user dataclass representation

    Not to be confused with the User database ORM model.  Logged in Auth User
    requires its own interface and abstraction from any model.
    """

    @property
    def name(self):
        """First and last name"""
        return str(self.first_name + ' ' + self.last_name).strip()

    @property
    def avatar_url(self):
        """Alias to avatar"""
        return self.avatar

    @property
    def admin(self):
        """Check if user is a superadmin"""
        return self.superadmin

    @property
    def is_admin(self):
        """Check if user is a superadmin"""
        return self.superadmin

    @property
    def is_superadmin(self):
        """Check if user is a superadmin"""
        return self.superadmin

    @property
    def is_not_admin(self):
        """Check if user is not a superadmin"""
        return not self.superadmin

    @property
    def is_authenticated(self):
        """Check if user is logged in"""
        return self.authenticated

    @property
    def loggedin(self):
        """Check if user is logged in"""
        return self.authenticated

    @property
    def is_loggedin(self):
        """Check if user is logged in"""
        return self.authenticated

    @property
    def is_not_loggedin(self):
        """Check if user is not logged in"""
        return not self.authenticated

    @property
    def is_not_authenticated(self):
        """Check if user is not logged in"""
        return not self.authenticated

    @property
    def check(self):
        """Alias to authenticated"""
        return self.authenticated

    def can(self, permissions: Union[str, List]) -> bool:
        """Check if user has ALL of these permissions"""
        if self.superadmin: return True

        permissions = [permissions] if isinstance(permissions, str) else list(permissions)

        for permission in permissions:
            if permission not in self.permissions:
                return False
        return True

    def can_any(self, permissions: Union[str, List]) -> bool:
        """Check if user has any one of these permissions"""
        if self.superadmin: return True

        permissions = [permissions] if isinstance(permissions, str) else list(permissions)

        for permission in permissions:
            if permission in self.permissions:
                return True
        return False

    def cant(self, permissions: Union[str, List]) -> bool:
        """Check if user does not have one of these permissions"""
        return not self.can(permissions)

    def cannot(self, permissions: Union[str, List]) -> bool:
        """Alias to cant"""
        return self.cant(permissions)






import uvicore
from uvicore.http.routing import Routes, ApiRouter, ModelRouter


@uvicore.routes()
class Api(Routes):

    # Apply scopes to all routes and children controllers
    #scopes = ['authenticated']
    #scopes = ['authenticated', 'employee']
    #scopes = None

    def register(self, route: ApiRouter):
        """Register API Route Endpoints"""

        # Define controller base path
        route.controllers = 'uvicore.auth.http.api'

        @route.group(tags=['Auth'])
        def private_routes():

            @route.get('/auth/userinfo')
            def userinfo():
                return {'hi': 'there'}

        # Return router
        return route



import json
import uvicore
import inspect
from uvicore import typing
from uvicore.http.routing.api_router import ApiRouter
from uvicore.http.routing.router import Routes as Controller
from uvicore.http.request import Request
from uvicore.support import str as string
from uvicore.support.dumper import dump, dd
from uvicore.support import module
from uvicore.http.exceptions import HTTPException, PermissionDenied
#from uvicore.auth import User
from uvicore.contracts import User
from uvicore.http.routing import Guard


class NestedRoute:

    def routes(self, Model, route, path, tags, scopes):
        """Build dynamic model CRUD routes"""



@uvicore.service()
class ModelRoute:
    """Dynamic Model CRUD Routes"""

    @property
    def log(self):
        return uvicore.log.name('uvicore.model_router')

    def routes(self, Model, route, path, tags, scopes):
        """Build dynamic model CRUD routes"""

        # Rest Notes
        """
        https://www.mscharhag.com/p/rest-api-design
        https://www.mscharhag.com/api-design/http-post-put-patch

        GET
        HEAD
        OPTIONS
        TRACE

        POST is for new records
            POST /spaces
            Not idempotent as it will continue to create new resources
            If new post, return 200
            If endpoint has no response but created, return 204 (no content but success)

        PUT is for updating existing records whos ID is in the url
            PUT /spaces/123
            Not for partial updates, expects the FULL object
            Idempotent

        PATCH
            PATCH /spaces/123
            Like PUT, but can be partial object, or could be full, either way
            Updates only records defined in the partial object

        DELETE
        """



        # URL query notes
        """
        Include
        -------
        include=sections.topics

        Where AND
        ----------
        where={"id": 1}
        where={"id": 1, "name": "test"}

        where={"id": [">", 1]}

        where={"id": ["in", ["one", "two"]]}
        where={"id": [">", 5], "name": "asdf", "email": ["like", "asdf"]}


        Where OR
        --------
        or_where=(id,1)+(id,3)

        Group By
        --------

        Order By
        --------
        order_by=[{"id": "ASC"}, {"name": "DESC"}]

        Paging
        ------
        page=1
        size=10
        translates to ORM limit and offset

        Cache
        -----
        cache=60  in seconds




        """


        def build_query(Model, *,
            include: str,
            where: str,
        ):
            query = Model.query()

            # Include
            if include: query.include(*include.split(','))

            # Where
            _build_where(query, where)


            return query

        def _build_where(query, where_str: str):
            if not where_str: return
            try:
                # Convert where string JSON to python object
                wheres = json.loads(where_str)
                #dump(where_str, wheres)

                # Where must be a dict
                if not isinstance(wheres, dict): return

                # WORKS - where={"id": [">", 5]}
                # WORKS - where={"id": ["in", [1, 2, 3]]}
                # WORKS - where={"title": ["like", "%black%"]}
                # WORKS - where={"id": 65, "topic_id": ["in", [1, 2, 3, 4, 5]]}
                # WORKS - ?include=topic.section.space&where={"topic.slug": "/tools", "topic.section.slug": "/apps", "topic.section.space.slug": "/dev"}

                for (key, value) in wheres.items():
                    #dump("Key: " + str(key) + " - Value: " + str(value))
                    if isinstance(value, typing.List):
                        if len(value) != 2: continue  # Valid advanced value must be 2 item List
                        operator = value[0]
                        value = value[1]
                        query.where(key, operator, value)
                        #dump("advanced: ", operator, value, '-----')
                    else:
                        query.where(key, value)

            except Exception as e:
                self.log.error(e)

        @route.get('/' + path, response_model=typing.List[Model], tags=tags, scopes=[Model.tablename + '.read'] if scopes is None else scopes)
        #async def list(include: typing.Optional[str] = '', user: User = Guard(Model.tablename + '.read')):
        async def list(
            request: Request,
            include: typing.Optional[str] = '',
            where: typing.Optional[str] = '',
        ):
            # The auth guard will not allow this method, but we do have to check any INCLUDES against that models permissions
            user: User = request.user
            includes = include.split(',') if include else []
            self.guard_include_permissions(Model, includes, user)

            # Build ORM query from URL query strings
            query = build_query(Model,
                include=include,
                where=where
            )

            # Run ORM query for results
            try:
                results = await query.get()
                return results
            except:
                raise HTTPException(500, str("Error in query builder, most likely an unknown column or query parameter."))

        @route.get('/' + path + '/{id}', response_model=Model, tags=tags, scopes=[Model.tablename + '.read'] if scopes is None else scopes)
        #async def get(id: typing.Any, include: typing.Optional[str] = '', user: User = Guard(Model.tablename + '.read')):
        async def get(request: Request, id: typing.Any, include: typing.Optional[str] = ''):
            # The auth guard will not allow this method, but we do have to check any INCLUDES against that models permissions
            user: User = request.user
            includes = include.split(',') if include else []
            self.guard_include_permissions(Model, includes, user)

            # Build ORM query from URL query strings
            query = build_query(Model,
                include=include,
                #where=where, NO not yet, with find
            )

            # Run ORM query for results
            #return await Model.query().include(*includes).cache().find(id)
            try:
                results = await query.find(id)
                return results
            except:
                raise HTTPException(500, str("Error in query builder, most likely an unknown column or query parameter."))


        @route.post('/' + path, response_model=Model, tags=tags, scopes=[Model.tablename + '.create'] if scopes is None else scopes)
        async def post(request: Request, item: Model):
            # NOTES
            # How do I check each child relations permissions, there is no includes
            # Would have to flip each item key and check if its a relation?
            # Then check if user has access to that relation?

            # Insert into storate and return primary key inserted
            try:
                result = await Model.insert(item)

                # If primary key is read_only, assume its an auto-incrementing pk
                # If not read_only, its a manual pk like 'key'.
                # Only set new pk result if pk is read_only.  Why? Because when pk is 'key'
                # a string, encode/databases does not return the new pk, just returns 1 every time.
                if Model.modelfield(Model.pk).read_only == True:
                    setattr(item, Model.pk, result)

                # Return inserted item
                return item

            except Exception as e:
                raise HTTPException(500, str(e))



    def guard_include_permissions(self, Model, includes: typing.List, user: User):
        # No includes, skip
        if not includes: return

        # User is superadmin, allow
        if user.superadmin: return

        # Loop each include and parts
        for include in includes:

            # Includes are split into dotnotation "parts".  We have to walk down each parts relations
            parts = [include]
            if '.' in include: parts = include.split('.')

            entity = Model
            for part in parts:

                # Get field for this "include part".  Remember entity here is not only Model, but a walkdown
                # based on each part (separated by dotnotation)
                field = entity.modelfields.get(part)  # Don't use modelfield() as it throws exception

                # Field not found, include string was a typo
                if not field: continue

                # Field has a column entry, which means its NOT a relation
                if field.column: continue

                # Field is not an actual relation
                if not field.relation: continue

                # Get actual relation object
                relation = field.relation.fill(field)

                # Get model permission string for this relation
                model_permissions = [relation.entity.tablename + '.read'] if scopes is None else scopes

                # User must have ALL scopes defined on the route (an AND statement)
                authorized = True
                missing_scopes = []
                for scope in model_permissions:
                    if scope not in user.permissions:
                        authorized = False
                        missing_scopes.append(scope)

                # # Check if user has permissino to child relatoin (or superadmin)
                # if model_permission not in user.permissions:
                #     # I convert model_permissins to a list for consistency with other permission denied errors
                #     # although there will always be just one model_permission in this function
                #     # raise HTTPException(
                #     #     status_code=401,
                #     #     detail="Permission denied to {}".format(str([model_permission]))
                #     # )
                #     print('be')
                #     raise PermissionDenied(model_permission)

                if not authorized:
                    raise PermissionDenied(missing_scopes)

                # Walk down each "include parts" entities
                entity = relation.entity



            # if not field: continue
            # if not field.relation: continue
            # relation: Relation = field.relation.fill(field)
            # tablename = relation.entity.tablename
            # permission = tablename + '.read'

            # # Check if user has permissino to child relatoin (or superadmin)
            # if user.superadmin == False and permission not in user.permissions:
            #     raise HTTPException(
            #         status_code=401,
            #         detail="Access denied to {}".format(tablename)
            #     )




        # @route.post('/' + path, response_model=Model, tags=tags)
        # async def post(entity: Model):
        #     return entity


        # # GET List
        # router.add_api_route(
        #     path='/' + path,
        #     methods=['GET'],
        #     tags=tags,
        #     endpoint=self.list,
        #     response_model=List[Model],
        #     name='Get all {}'.format(path),
        #     #name=path,
        #     #   name='',
        # )

        # # GET
        # router.add_api_route(
        #     path='/' + path + '/{id}',
        #     methods=['GET'],
        #     tags=tags,
        #     endpoint=self.get,
        #     response_model=Model,
        #     name='Get {} by ID'.format(path),
        #     #name=path,
        #     #name='',
        # )

        # # POST
        # router.add_api_route(
        #     path='/' + path,
        #     methods=['POST'],
        #     tags=tags,
        #     endpoint=self.post,
        #     response_model=Model,
        #     name='Create {}'.format(path),
        #     #name=path,
        #     #name='',
        # )

    # async def list(self, include: str = ''):
    #     return await Model.query().include(*include.split(',')).get()

    # async def get(self, id: Any, include: str = ''):
    #     return await Model.query().include(*include.split(',')).find(id)

    # async def post(self, entity: Model):
    #     #dump(request.__dict__)
    #     return entity
    #     #return {'name': 'hi'}
    #     #return await Model.query().include(*include.split(',')).find(id)


@uvicore.controller()
class ModelRouter(Controller):

    def __init__(self, package, scopes: typing.List = None):
        """Init override to add custom guards parameter"""
        super().__init__(package)
        self.scopes = scopes

    def register(self, route: ApiRouter):

        # Get all models in tablename sorted order from Ioc bindings
        models = self.models()

        # Get source code for ModelRouter
        modelroute = inspect.getsource(ModelRoute)
        modelroute += "ModelRoute().routes(Model, route, path, tags, scopes)"

        #from uvicore.auth.middleware import Guard
        #from uvicore.auth.models import User


        # Loop each sorted model and dynamically add a ModelRouter
        for key, binding in models.items():
            # Temp, only hashtag
            #if binding.object.tablename != 'hashtags': continue

            # Get model information from Ioc binding
            Model = binding.object
            modelname = binding.path.split('.')[-1]
            tablename = Model.tablename
            path = tablename
            tags = [string.ucbreakup(path)]
            permissions = Model.tablename

            # Dynamically instantiate ModelRouter from source code and exec
            # passing in the proper globals.  Why?  Why not just instantiate
            # the class and be done?  Becuase pydantic doesn't understand
            # type hinting that way.  For example post(entity: Model) does NOT
            # work as the type hinter errors on the Model.  I tried every other way
            # and type hinting just does not work as expected.  Even with Generics[E]
            # Not sure why exactly. This dynamic execution however works for now.
            # One caveat to this is you cannot override and EXTEND the ModelRoute
            # class.  You can override, but you must re-impliment the entire class
            # and not just extend it.
            exec(modelroute, {
                'Model': Model,
                'route': route,
                'path': path,
                'tags': tags,
                'uvicore': uvicore,
                'typing': typing,
                'User': User,
                'Guard': Guard,
                'Request': Request,
                'HTTPException': HTTPException,
                'PermissionDenied': PermissionDenied,
                'scopes': self.scopes,
                'dump': dump,
                'json': json,
            })

        # Return router
        return route

    def models(self) -> typing.OrderedDict:
        """Get all models in tablename sorted order from Ioc bindings"""
        models = typing.OrderedDict()
        unsorted_models = {}
        model_bindings = uvicore.ioc.binding(type='model', include_overrides=False)
        for key, binding in model_bindings.items():
            tablename = binding.object.tablename
            if tablename not in models:
                unsorted_models[tablename] = binding

        # Sort the OrderedDict
        #import operator
        #x = sorted(models.items(), key=lambda x: x[1])
        #x = sorted((value, key) for (key,value) in models.items())
        keys = sorted(unsorted_models.keys())
        for key in keys:
            models[key] = unsorted_models[key]
        return models
























#             list_method = """
# @router.get('/' + path, response_model=List[Model], tags=tags)
# async def list(include: str = ''):
#     return await Model.query().include(*include.split(',')).get()
# """

#             get_method = """
# @router.get('/' + path + '/{id}', response_model=Model, tags=tags)
# async def get(id: Any, include: str = ''):
#     return await Model.query().include(*include.split(',')).find(id)
# """

#             post_method = """
# @router.post('/' + path, response_model=Model, tags=tags)
# async def post(entity: Model):
#     return entity
# """

#             # Add dynamic methods
#             exec(list_method, exec_globals)
#             exec(get_method, exec_globals)
#             exec(post_method, exec_globals)



import uvicore
from uvicore.console import click, group
from uvicore.console.events import command as ConsoleEvents

title_ideas = """
    A Fast Async Python Framework for CLI, Web and API

    High Performance Web, API and CLI Python Framework

    Python async Web, API and CLI fullstack Framework

    Fullstack Asynchronous Python Framework
    Python web framework
    Python Fullstack Asynchronous Web, API and CLI Framework

    Asynchronous Python Web, API and CLI Fullstack Framework

    The Async Python Framework for Craftsmen
    The Async Python Framework for Artisans
    The Artisanal Asynchronous Python Framework
"""

# Must include actual bind name when stacking decorators as it cannot deduce on its own.
# Bind decorator must come first in the stack
@uvicore.service('uvicore.console.console.cli', aliases=['Console', 'console', 'cli'])
@group(help=f"""
    \b
    Uvicore {uvicore.__version__}
    The Async Python Framework for Artisans.
    An Elegant Fullstack Python Web, API and CLI Framework

    Copyright (c) 2020 Matthew Reschke
    License http://mreschke.com/license/mit
""")
@click.version_option(version=uvicore.__version__, prog_name='Uvicore Framework', flag_value='--d')
@click.pass_context
async def cli(ctx):
    # Console startup handler
    # This runs before any command.  Even if you don't actually run a command but only display
    # the help output using NO parameters, like ./uvicore http  It does not however run when just
    # showing ./uvicore main help output.  Only on subcommand help output or actual command
    await console_startup()

    # Console shutdown handler
    ctx.call_on_close(console_shutdown)

    # The actual command runs right here


# Had to hack asyncclick core.py def close and __aexit__ with awaits
# github issue open, hopefully he will fix soon


async def console_startup():
    #print('console startup')
    await ConsoleEvents.Startup().codispatch()


#@cli.resultcallback()
#async def console_shutdown(result, **kwargs):
#This decorator works only if you give it an actual command
#Just showing help output from NO command does not fire this, so aiohttp still complains
#about not being shutdown properly even if you just do ./uvicore http with no actual command to run.

async def console_shutdown():
    #print('console shutdown')
    await ConsoleEvents.Shutdown().codispatch()



import uvicore
from uvicore.support.dumper import dump, dd
from uvicore.console import command, argument, option


@command()
async def cli():
    """xx_AppName xx_name CLI command"""

    print('Welcome to your new xx_AppName xx_name CLI command!')



import uvicore
from uvicore.typing import Any, Dict
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.support.module import load
from uvicore.console.provider import Cli
from uvicore.database import bootstrap
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Database(ServiceProvider, Cli):

    def register(self) -> None:
        # Register IoC bindings
        # self.bind('Database', 'uvicore.database.db._Db',
        #     singleton=True,
        #     aliases=['database', 'db'],
        # )

        # Set uvicore.log global
        uvicore.db = uvicore.ioc.make('uvicore.database.db.Db')

        # Register event listeners
        AppEvents.Booted.listen(bootstrap.Database)

        # String based events instead of class based because HTTP may not even
        # be installed, so importing it would cause an issue.
        @uvicore.events.handle(['uvicore.console.events.command.Shutdown', 'uvicore.http.events.server.Shutdown'])
        async def uvicore_shutdown(event):
            # Disconnect from all connected databases
            await uvicore.db.disconnect(from_all=True)

    def boot(self) -> None:
        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define db commands
        self.commands(
            group='db',
            help='Database Commands',
            commands={
                'create': 'uvicore.database.commands.db.create',
                'drop': 'uvicore.database.commands.db.drop',
                'recreate': 'uvicore.database.commands.db.recreate',
                'seed': 'uvicore.database.commands.db.seed',
                'reseed': 'uvicore.database.commands.db.reseed',
                'connections': 'uvicore.database.commands.db.connections',
            }
        )

        # Extend schematic generator commands
        self.commands(
            group='gen',
            commands={
                'table': 'uvicore.database.commands.generators.table',
                'seeder': 'uvicore.database.commands.generators.seeder',
            }
        )



import uvicore
from jwt import decode
from uvicore.support.dumper import dump, dd
from uvicore.http.request import HTTPConnection
from uvicore.typing import Dict, Optional, Union, Callable
from uvicore.auth.authenticators.base import Authenticator
from uvicore.http.exceptions import NotAuthenticated, InvalidCredentials, HTTPException
from uvicore.contracts import User

@uvicore.service()
class Jwt(Authenticator):
    """JWT bearer token authenticator"""

    # Notice:  Do not ever throw or return an error from authenticators.
    # If there is any auth problem (no headers, invalid or expired tokens, bad password)
    # then return True or False instead.  Any issues or invalid credentials causes the global
    # Authentication middleware to inject an Anonymous user, not to throw errors.
    # Any permissions errors happen at a route level when checking of the route
    # require an authenticated user or valid scope/role.

    # Return of False means this authentication method is not being attempted, try next authenticator
    # Return of True means this authentication method was being attempted, but failed validation, skip next authenticator
    # Return of User object means a valid user was found, skip next authenticator

    async def authenticate(self, request: HTTPConnection) -> Union[User, bool]:
        #dump('JWT Authenticator HERE')

        # Parse authorization header
        authorization, scheme, token = self.auth_header(request)

        # This authentication method not provided or attempted, goto next authenticator
        if not authorization or scheme != "bearer":
            # Return of False means this authentication method is not being attempted
            # goto next authenticator in stack
            return False

        # Detect anonymous API access.
        # API gateways that allow anonymous access to an API will pass the request
        # downstream even if the token is missing, invalid or expired.
        # When they do, they add a header notifying downstream that auth has failed
        # but anonymous access is still allowed.  For Kong this header is x-anonymous-consumer
        # If the API gateway does not allow anonymous access, downstream won't even be proxied.
        #dump(request.headers)
        if (self.config.anonymous_header and self.config.anonymous_header in request.headers):
            # Anonymous header found, return True to denote Anonymous user and skip next authenticator
            return True

        # Decode JWT with or without verification
        jwt = None
        try:
            if self.config.verify_signature:
                # With secret algorithm validation
                #dump('WITH Validation')
                jwt = Dict(decode(token, self.config.secret, audience=self.config.audience, algorithms=self.config.algorithms))
            else:
                # Without validation
                #dump('WITHOUT Validation')
                jwt = Dict(decode(token, options={"verify_signature": False}))

                # Validate aud claim
                # jwt.decode also validates the aud claim.  Since we are skipping validation, we'll still validate aud claim here.
                if jwt.aud != self.config.audience:
                    # Audience mismatch, return True to denote Anonymous user and skip next authenticator
                    return True
        except Exception as e:
            # Issue with validating JWT, return True to denote Anonymous user and skip next authenticator
            return True

        #dump('JWT', jwt)

        # Get user and validate credentials
        user: User = await self.retrieve_user(jwt.email, None, self.config.provider, request, jwt=jwt)

        # User from valid JWT not found in uvicore OR not synced for first time (no uuid).
        # Auto create or update user if allowed in config
        if self.config.auto_create_user and (user is None or user.uuid is None):
            jwt_mapping = self.config.auto_create_user_jwt_mapping
            new_user = {}
            for key, value in self.config.auto_create_user_jwt_mapping.items():
                if isinstance(value, Callable):
                    new_user[key] = value(jwt)
                else:
                    new_user[key] = value

            # User does not exist, create user
            if user is None:
                # Auto create new user in user provider
                await self.create_user(self.config.provider, request, **new_user)

            else:
                # User exists, but needs an initial sync
                await self.sync_user(self.config.provider, request, **new_user)

            # Get user and validate credentials
            user: User = await self.retrieve_user(jwt.email, None, self.config.provider, request, jwt=jwt)


            # Dict({
            #     'aud': 'd709a432-5edc-44c7-8721-4cf123473c45',
            #     'exp': 1616000280,
            #     'iat': 1615996680,
            #     'iss': 'https://auth-local.sunfinity.com',
            #     'sub': '217e27b4-0e0a-464c-84a8-c40312b55801',
            #     'authenticationType': 'PASSWORD',
            #     'email': 'it@sunfinity.com',
            #     'email_verified': True,
            #     'applicationId': 'd709a432-5edc-44c7-8721-4cf123473c45',
            #     'roles': ['Administrator'],
            #     'name': 'Admin|Istrator'
            # })


        # If user is none and auto_create_user is enabled, auto-create user
        # Link user up to groups table based on JWT roles
        # Now self.get_user again

        # If no user returned, validation has failed or user not found
        #if user is None: raise InvalidCredentials()

        # Validate Permissions
        #self.validate_permissions(user, scopes.scopes)

        # Authentication successful.
        # Add user to request in case we use it in a decorator, we can pull it out with request.scope.get('user')
        #request.scope['user'] = user

        # Return user.  If no user return True to denote Anonymous User and skip next authenticator
        return user or True



import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.redis import bootstrap
from uvicore.foundation.events import app as AppEvents
from uvicore.redis.provider import Redis as RedisMixin


@uvicore.provider()
class Redis(ServiceProvider, RedisMixin):

    def register(self) -> None:
        # Register event listeners
        AppEvents.Booted.listen(bootstrap.Redis)


    def boot(self) -> None:
        pass
        # Define service provider registration control
        #self.registers(self.package.config.registers)

        # Define Redis Connections
        # self.redis_connections(
        #     connections=self.package.config.redis.connections,
        #     default=self.package.config.redis.default
        # )



from uvicore import app
import typing
from databases import DatabaseURL
from sqlalchemy import create_engine
from typing import List, Mapping, TypeVar

E = TypeVar('E')

class Database:

    def __init__(self, url: str):
        #self.url = DatabaseURL(url)
        self.url = url
        if app.is_async:
            from databases import Database as EncodeDatabase
            # NO +pymysql
            self.encodedb = EncodeDatabase("mysql://root:techie@127.0.0.1/uvicore_wiki")
            @app.http.on_event("startup")
            async def startup():
                await self.encodedb.connect()

            @app.http.on_event("shutdown")
            async def shutdown():
                await self.encodedb.disconnect()
        else:

            self.engine = create_engine(self.url, echo=True)

        from sqlalchemy import MetaData
        self.metadata = MetaData()

    def connect(self):
        return self.engine.connect()

    def fetchone(self, entity: E, query) -> E:
        if app.is_async:
            return self.encodedb.fetch_one(query=query)
        else:
            with self.connect() as con:
                results = con.execute(query)
                return entity(**results.fetchone())

    def fetchall(self, entity, query) -> List[E]:
        if app.is_async:
            return self.encodedb.fetch_all(query=query)
        else:
            with self.connect() as con:
                results = con.execute(query)
                rows = []
                for row in results:
                    #rows.append(entity.User(**row))
                    rows.append(entity(**row))
                return rows




from __future__ import annotations

import uvicore
from typing import Optional
from uvicore.auth.models.user import User
from uvicore.support.dumper import dd, dump
from uvicore.orm import Model, ModelMetaclass, BelongsTo, Field
from xx_vendor.xx_appname.database.tables import xx_tablename as table


@uvicore.model()
class xx_ModelName(Model['xx_ModelName'], metaclass=ModelMetaclass):
    """xx_AppName xx_ModelName Model"""

    # Database table definition
    __tableclass__ = table.xx_TableName

    id: Optional[int] = Field('id',
        primary=True,
        description='Post ID',
        #sortable=True,
        #searchable=True,
        read_only=True,
    )

    slug: str = Field('slug',
        description='URL Friendly Post Title Slug',
        required=True,
    )

    title: str = Field('title',
        description='Post Title',
        required=True,
    )



# Public API used in packages routes and controllers
from .api_router import ApiRoute, ApiRouter
from .guard import Guard
from .model_router import ModelRouter
from .router import Router
from .router import Routes
from .router import Routes as Controller
from .web_router import WebRoute, WebRouter



from uvicore.typing import OrderedDict

# This is the main foundation config.  All items here can be overridden
# when used inside other applications.  Accessible at config('uvicore.foundation')

config = {

    # --------------------------------------------------------------------------
    # Registration Control
    # --------------------------------------------------------------------------
    # This lets you control the service provider registrations.  If this app
    # is used as a package inside another app you might not want some things
    # registered in that context.
    # 'registers': {
    #     'web_routes': False,
    #     'api_routes': False,
    #     'middleware': False,
    #     'views': False,
    #     'assets': False,
    #     'commands': True,
    #     'models': False,
    #     'tables': False,
    #     'seeders': False,
    # },


    'commands': {
        # Register Ioc commands
        'ioc': {
            'help': 'Uvicore Ioc (Inversion of Control) Information',
            'commands': {
                'bindings': 'uvicore.container.commands.ioc.bindings',
                'singletons': 'uvicore.container.commands.ioc.singletons',
                'overrides': 'uvicore.container.commands.ioc.overrides',
                'type': 'uvicore.container.commands.ioc.type',
                'get': 'uvicore.container.commands.ioc.get',
            },
        },

        # Register Package commands
        'package': {
            'help': 'Uvicore Package Information',
            'commands': {
                'providers': 'uvicore.package.commands.package.providers',
                'list': 'uvicore.package.commands.package.list',
                'get': 'uvicore.package.commands.package.get',
            },
        },

        # Register Event commands
        'event': {
            'help': 'Uvicore Event Information',
            'commands': {
                'list': 'uvicore.events.commands.event.list',
                'get': 'uvicore.events.commands.event.get',
            },
        },
    },


    # --------------------------------------------------------------------------
    # Package Dependencies (Service Providers)
    #
    # Define all the packages that this package depends on.  At a minimum, only
    # the uvicore.foundation package is required.  The foundation is very
    # minimal and only depends on configuratino, logging and console itself.
    # You must add other core services built into uvicore only if your package
    # requires them.  Services like uvicore.database, uvicore.orm, uvicore.http
    # uvicore.auth...
    # --------------------------------------------------------------------------
    'dependencies': OrderedDict({
        # Careful to only define REQUIRED services for a minimal Uvicore.  Things
        # like auth, db, http, etc... are optional and specified in a packages
        # dependencies if needed.

        # Configuration is used right away in provider register()
        'uvicore.configuration': {
            'provider': 'uvicore.configuration.services.Configuration',
        },

        # I want logging quick as it may be used in providers register() and boot()
        'uvicore.logging': {
            'provider': 'uvicore.logging.services.Logging',
        },

        # Console is required for every uvicore app
        'uvicore.console': {
            'provider': 'uvicore.console.services.Console',
        },

        # Cache is required for every uvicore app
        'uvicore.cache': {
            'provider': 'uvicore.cache.services.Cache',
        },
    }),

}



import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.console.provider import Cli

@uvicore.provider()
class Configuration(ServiceProvider, Cli):

    def register(self) -> None:
        """Register package into uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs and IoC bindings.  Configs are deep merged only after
        all packages are registered.  No real work should be performed here as it
        is very early in the bootstraping process and most internal processes are not
        instantiated yet."""

        # Register IoC bindings
        # self.bind('Configuration', 'uvicore.configuration.configuration._Configuration',
        #     aliases=['Config', 'config'],
        #     singleton=True,
        # )

        # Register configuration system
        self.register_configuration()

    def boot(self) -> None:
        """Bootstrap package into uvicore framework.
        Boot takes place after all packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all configs.
        This is where you load views, assets, routes, commands..."""

        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs

        # Define commands
        self.commands(
            group='config',
            help='Configuration Information',
            commands={
                'list': 'uvicore.configuration.commands.config.list',
                'get': 'uvicore.configuration.commands.config.get',
            }
        )

    def register_configuration(self) -> None:
        # Set uvicore.log global
        uvicore.config = uvicore.ioc.make('uvicore.configuration.configuration.Configuration')

        # Set app.config for convenience (only after register since config is a service provider itself)
        # No, don't want duplicate entry points everywhere
        #self.app._config = uvicore.config

        # Set main app config
        uvicore.config.app = self.app_config



from uvicore.typing import Sequence, List
from fastapi.security import SecurityScopes
from starlette.requests import Request, HTTPConnection
from fastapi.params import Path, Query, Header, Cookie, Body, Form, File, Depends, Security
from uvicore.support.dumper import dump, dd
import json

class Parameter(Security):
    """Base class for user defined request parameters with infinite kwargs!"""

    # Parameters can accept any number of kwargs which are hacked as
    # a single "string" security scope converted to json then converted
    # back to an object before the handler is called

    #def __init__(self, scopes: Sequence[str] = []):
    def __init__(self, **kwargs):
        # Hack, scopes must be strings, convert kwargs into one JSON string
        scopes = [json.dumps(kwargs)]

        # Call FastAPI Security Depends with "self" and one scope as a JSON string
        super().__init__(dependency=self, scopes=scopes, use_cache=True)

    async def __call__(self, security_scopes: SecurityScopes, request: Request):
        # Hack, scopes are a JSON blog of kwargs
        kwargs = json.loads(security_scopes.scopes[0])

        # Call user defined handler
        return await self.handle(request=request, **kwargs)



import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.console.provider import Cli


@uvicore.provider()
class Foundation(ServiceProvider, Cli):

    def register(self):

        # Register configs
        self.configs([
            {'key': self.name, 'module': 'uvicore.foundation.config.package.config'}
        ])

        # Register events used in this package
        # self.events.register(
        #     name='uvicore.foundation.events.app.Registered',
        #     description='Application bootstrap has registered all package service providers.',
        #     dynamic=False,
        #     is_async=False,
        # )
        # self.events.register(
        #     name='uvicore.foundation.events.app.Booted',
        #     description='Application bootstrap has booted all package service providers.',
        #     dynamic=False,
        #     is_async=False,
        # )

        # # Register root level commands
        # self.commands(
        #     name='root',
        #     #help='xx',
        #     commands=[
        #         ('version', 'uvicore.foundation.commands.version.cli'),
        #     ]
        # )

    def boot(self):
        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs

        # Define commands
        # Why here in foundation?  Because these items are not actual packages
        # with their own service providers to register themselves.
        # self.commands({
        #     # Register Ioc commands
        #     'ioc': {
        #         'help': 'Uvicore Ioc (Inversion of Control) Information',
        #         'commands': {
        #             'bindings': 'uvicore.container.commands.ioc.bindings',
        #             'singletons': 'uvicore.container.commands.ioc.singletons',
        #             'overrides': 'uvicore.container.commands.ioc.overrides',
        #             'type': 'uvicore.container.commands.ioc.type',
        #             'get': 'uvicore.container.commands.ioc.get',
        #         },
        #     },

        #     # Register Package commands
        #     'package': {
        #         'help': 'Uvicore Package Information',
        #         'commands': {
        #             'providers': 'uvicore.package.commands.package.providers',
        #             'list': 'uvicore.package.commands.package.list',
        #             'get': 'uvicore.package.commands.package.get',
        #         },
        #     },

        #     # Register Event commands
        #     'event': {
        #         'help': 'Uvicore Event Information',
        #         'commands': {
        #             'list': 'uvicore.events.commands.event.list',
        #             'get': 'uvicore.events.commands.event.get',
        #         },
        #     },
        # })

        # Alternative in kwargs format

        # # Register Ioc commands
        # self.commands(
        #     group='ioc',
        #     help='Uvicore Ioc (Inversion of Control) Information',
        #     commands={
        #         'bindings': 'uvicore.container.commands.ioc.bindings',
        #         'singletons': 'uvicore.container.commands.ioc.singletons',
        #         'overrides': 'uvicore.container.commands.ioc.overrides',
        #         'type': 'uvicore.container.commands.ioc.type',
        #         'get': 'uvicore.container.commands.ioc.get',
        #     }
        # )

        # # Register Package commands
        # self.commands(
        #     group='package',
        #     help='Uvicore Package Information',
        #     commands={
        #         'providers': 'uvicore.package.commands.package.providers',
        #         'list': 'uvicore.package.commands.package.list',
        #         'get': 'uvicore.package.commands.package.get',
        #     }
        # )

        # # Register Event commands
        # self.commands(
        #     group='event',
        #     help='Uvicore Event Information',
        #     commands={
        #         'list': 'uvicore.events.commands.event.list',
        #         'get': 'uvicore.events.commands.event.get',
        #     },
        # )

        # Alternative, pull from a config
        self.commands(self.package.config.commands)



import uvicore
from uvicore.auth.user import User
from uvicore.support.hash import sha1
from uvicore.contracts import UserProvider
from uvicore.support.dumper import dump, dd
from uvicore.auth.support import password as pwd
from uvicore.http.request import HTTPConnection
from uvicore.typing import List, Union, Any, Dict
from uvicore.auth.models.user import User as Model
from uvicore.auth.models.group import Group
from datetime import datetime


@uvicore.service()
class Orm(UserProvider):
    """Retrieve and validate user from uvicore.auth ORM User model during Authentication middleware

    This is NOT a stateless user provider as it queries the user, groups, roles tables from a database.
    """

    # def __init__(self):
    #     # Only need for an __init__ override is to modify field mappings
    #     super().__init__()

    #     # Temp, until I add username to ORM model
    #     self.field_map['username'] = 'email'

    async def _retrieve_user(self,
        key_name: str,
        key_value: Any,
        request: HTTPConnection,
        *,
        password: str = None,

        # Parameters from auth config
        anonymous: bool = False,
        includes: List = None,

        # Must have kwargs for infinite allowed optional params, even if not used.
        **kwargs,
    ) -> User:
        """Retrieve user from backend"""

        # Get password hash for cache key.  Password is still required to pull the right cache key
        # or else someone could login with an invalid password for the duration of the cache
        password_hash = '/' + sha1(password) if password is not None else ''

        # Check if user already validated in cache
        cache_key = 'auth/user/' + str(key_value) + password_hash
        if await uvicore.cache.has(cache_key):
            # User is already validated and cached
            # Retrieve user from cache, no password check required because cache key has password has in it
            user = await uvicore.cache.get(cache_key)
            return user

        # ORM is currently thworing a Warning: Truncated incorrect DOUBLE value: '='
        # when using actual bool as bit value.  So I convert to '1' or '0' strings instead
        disabled = '1' if anonymous else '0'

        # Cache not found.  Query user, validate password and convert to user class
        find_kwargs = {key_name: key_value}
        db_user = await (Model.query()
            .include(*includes)
            .where('disabled', disabled)
            #.show_writeonly(['password'])
            .show_writeonly(True)
            .find(**find_kwargs)
        )

        # User not found or disabled.  Return None means not verified or found.
        if not db_user: return None

        # If we are checking passwords and the db_user has NO password, user cannot be logged into
        if password is not None and db_user.password is None: return None

        # If password, validate credentials
        if password is not None:
            if not pwd.verify(password, db_user.password):
                # Invalid password.  Return None means not verified or found.
                return None

        # Get users groups->roles->permissions (roles linked to a group)
        groups = []
        roles = []
        permissions = []
        if 'groups' in includes:
            user_groups = db_user.groups
            if user_groups:
                for group in user_groups:
                    groups.append(group.name)
                    if not group.roles: continue
                    for role in group.roles:
                        roles.append(role.name)
                        if not role.permissions: continue
                        for permission in role.permissions:
                            permissions.append(permission.name)

        # Get users roles->permissions (roles linked directly to the user)
        if 'roles' in includes:
            user_roles = db_user.roles
            if user_roles:
                for role in user_roles:
                    roles.append(role.name)
                    if not role.permissions: continue
                    for permission in role.permissions:
                        permissions.append(permission.name)

        # Unique groups, roles and permissions (sets are unique)
        groups = sorted(list(set(groups)))
        roles = sorted(list(set(roles)))
        permissions = sorted(list(set(permissions)))

        # Set super admin, existence of 'admin' permission
        # Fixme, there is a 'superadmin' field on the roles table.
        # If user is in any role with superadmin=True they are a superadmin
        superadmin = False
        if 'admin' in permissions:
            # No need for any permissinos besides ['admin']
            permissions = ['admin']
            superadmin = True

        # Build UserInfo dataclass with REQUIRED fields
        user = User(
            id=db_user.id,
            uuid=db_user.uuid,
            username=db_user.username,
            email=db_user.email,
            first_name=db_user.first_name,
            last_name=db_user.last_name,
            title=db_user.title,
            avatar=db_user.avatar_url,
            groups=groups,
            roles=roles,
            permissions=permissions,
            superadmin=superadmin,
            authenticated=not anonymous,
        )

        # Save to cache
        await uvicore.cache.put(cache_key, user)

        # Return to user
        return user

    async def create_user(self, request: HTTPConnection, **kwargs):
        """Create new user in backend"""
        # Pop groups from kwargs
        groups = kwargs.pop('groups')

        # Set other kwargs values
        kwargs['disabled'] = False
        kwargs['login_at'] = datetime.now()

        # Translate avatar
        kwargs['avatar_url'] = kwargs.pop('avatar')

        # Build user model
        user = Model(**kwargs)

        # Get actual groups in backend from groups array
        real_groups = await Group.query().where('name', 'in', groups).get()

        # Save user
        await user.save()

        # Link real_groups
        await user.link('groups', real_groups)

        # Return new backend user (not actual Auth user class)
        return user

    async def sync_user(self, request: HTTPConnection, **kwargs):
        # Get username
        username = kwargs['username']

        # Get actual backend user
        user = await Model.query().show_writeonly(['password']).find(username=username)

        # If we have successfully logged in, we are not disabled
        user.disabled = False
        user.login_at = datetime.now()

        # Pop groups from kwargs
        groups = kwargs.pop('groups')

        # Remove other kwargs items
        del kwargs['creator_id']

        # Translate avatar
        kwargs['avatar_url'] = kwargs.pop('avatar')

        # Add all other kwargs to user
        for key, value in kwargs.items():
            setattr(user, key, value)

        # Save user
        await user.save()

        # Return new backend user (not actual Auth user class)
        return user



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uvicore',
 'uvicore.auth',
 'uvicore.auth.authenticators',
 'uvicore.auth.commands',
 'uvicore.auth.config',
 'uvicore.auth.contractsOLD',
 'uvicore.auth.database.seeders',
 'uvicore.auth.database.tables',
 'uvicore.auth.http.api',
 'uvicore.auth.http.routes',
 'uvicore.auth.middleware',
 'uvicore.auth.models',
 'uvicore.auth.support',
 'uvicore.auth.user_providers',
 'uvicore.cache',
 'uvicore.cache.backends',
 'uvicore.configuration',
 'uvicore.configuration.commands',
 'uvicore.console',
 'uvicore.console.asyncclick',
 'uvicore.console.commands',
 'uvicore.console.commands.stubs',
 'uvicore.console.events',
 'uvicore.container',
 'uvicore.container.commands',
 'uvicore.contracts',
 'uvicore.database',
 'uvicore.database.OBSOLETE',
 'uvicore.database.commands',
 'uvicore.database.commands.stubs',
 'uvicore.events',
 'uvicore.events.commands',
 'uvicore.factories',
 'uvicore.foundation',
 'uvicore.foundation.config',
 'uvicore.foundation.decorators',
 'uvicore.foundation.events',
 'uvicore.http',
 'uvicore.http.OBSOLETE',
 'uvicore.http.commands',
 'uvicore.http.controllers',
 'uvicore.http.events',
 'uvicore.http.middleware',
 'uvicore.http.routing',
 'uvicore.http.servers',
 'uvicore.http.templating',
 'uvicore.http_client',
 'uvicore.logging',
 'uvicore.mail',
 'uvicore.mail.backends',
 'uvicore.orm',
 'uvicore.orm.commands',
 'uvicore.orm.commands.stubs',
 'uvicore.orm.drivers',
 'uvicore.package',
 'uvicore.package.commands',
 'uvicore.redis',
 'uvicore.support',
 'uvicore.typing']

package_data = \
{'': ['*'], 'uvicore.auth': ['http/public/*', 'http/public/assets/js/*']}

install_requires = \
['aiohttp>=3.7.0,<3.8.0',
 'anyio>=3.2.0,<3.3.0',
 'argon2-cffi>=20.1.0,<20.2.0',
 'colored>=1.4.0,<1.5.0',
 'cryptography>=3.4.0,<3.5.0',
 'environs>=9.3.0,<9.4.0',
 'prettyprinter>=0.18.0,<0.19.0']

extras_require = \
{'database': ['sqlalchemy>=1.3.0,<1.4.0',
              'alembic>=1.5.0,<1.6.0',
              'databases[postgresql,sqlite,mysql]>=0.4.0,<0.5.0'],
 'redis': ['aioredis>=1.3.0,<1.4.0'],
 'web': ['fastapi>=0.63.0,<0.64.0',
         'uvicorn>=0.13.0,<0.14.0',
         'gunicorn>=20.0.0,<20.1.0',
         'aiofiles>=0.6.0,<0.7.0',
         'Jinja2>=2.11.0,<2.12.0',
         'requests>=2.25.0,<2.26.0',
         'itsdangerous>=1.1.0,<1.2.0',
         'uvloop>=0.14.0,<0.15.0',
         'httptools>=0.1.0,<0.2.0',
         'python-multipart>=0.0.0,<0.1.0',
         'PyJWT>=2.0.0,<2.1.0']}

setup_kwargs = {
    'name': 'uvicore',
    'version': '0.1.15',
    'description': 'The Async Python Framework for Artisans. An Elegant Fullstack Python Web, API and CLI Framework',
    'long_description': '# Uvicore\n\nWelcome to Uvicore!  The Async Python Framework for Artisans. An Elegant Fullstack Python Web, API and CLI Framework.\n\n\n## Warning\n\nThis project is currently under heavy **active** development.  Although I am using this framework in Production personally, I am constantly building out the framework as I do. I would advise you against using it just yet.  It is mostly stable but **not feature complete**.\n\n\n## What is Uvicore?\n\n* A fullstack blazing fast asynchronous python Web, API and CLI framework.\n* Inspired by the elegance and design quality of Laravel.  Though it is NOT a Laravel clone by any stretch of the imagination.  If you want Laravel, go get Laravel and stop complaining about PHP :)\n* Not a mirco framework, up and running with a complete system in minutes, no boilerplate.\n* Deeply config driven with extremely deep merged and overridable package configs.\n* Package based.  Apps are packages, packages are apps.  No "shell" or "project" required.  Run it or use it in another running app, all in one.\n* Service Provider style package bootstrapping.\n* Full async database query builder built on encode/databases...or use the ORM!\n* Elegant custom async ORM inspired by Eloquent (but still very different).  Not another Django ORM clone thank God!\n* Complex and complete ORM relations including polymorphism.\n* Easy redis caching system (great for optimizing API database hits).\n* IoC (Inversion of Control).  Framework looks back to your app to determine import of every module.  You can override ANYTHING from core or other developers packages!\n* Blazing fast DUAL routers for Web and API usage, built on Starlette and FastAPI.\n* Perfectly suited for old-school server-side traditional post-back template driven web apps.\n* Perfectly suited for API backend only, quicker and easier that Django Rest Framework, built on FastAPI but with fullstack routing, controllers, ORM, caching...no boilerplate.\n* Websockets!  Well yeah, its async!\n* Automatic ORM Crud API\'s out of the box.  Never write a single controller or endpoint again.  Simply build Models. Instant API with granular roles and permission management.\n* ORM Models can query local database OR switch to remote API mode and query the data over uvicore automatic ORM crud APIs without code changing, syntax identical.  What?  This means your package IS the "server" and also the "client/sdk for API access".  Let that sink in a bit.\n* Granular groups, roles, permission management with Authentication middleware and route guards.\n* Quick and easy JWT authentication with an external IDP (FusionAuth!)\n* Multiple middleware authenticators.  APIs can autodetect and respond to Basic Auth, Digest, JWT and more.\n* Automatic OpenAPI 3.0 generated documentation for all API routes include automatic model router.\n* Fully async CLI based on asyncclick.  Query async ORM models and jobs from the CLI exactly as you would from an async web controller.  Everything can call async methods!\n* Event pub/sub system.  Subscribe to all Uvicore core events in the bootstrap process, ORM CRUD, authentication systems and any event fired from your packages.\n* And much more....documentation to come soon!\n\n\n\n## Feeling Dangerous?\n\nThis `uvicore/framework` repo is NOT what you clone if you want to run uvicore.  This is the framework.  The pypi package your actual app will depend on.  The `uvicore-installer` will create the folder structure for your app and include `uvicore/framework` as a dependency.  If you wanted to see what an app structure looks like, see https://github.com/uvicore/app which is what the `uvicore-installer` clones and regexes to your configured needs.  Its a package schematic.  **Use the installer.**\n\nIt works with a full installer, but there is little documentation yet so its up to you to deduce the system, though it\'s reasonably self explanatory.\n\n**Installation**\n```bash\nwget https://raw.githubusercontent.com/uvicore/framework/master/bin/uvicore-installer\nmv uvicore-installer /usr/local/bin\nchmod a+x /usr/local/bin/uvicore-installer\n\ncd ~/Code\nuvicore-installer  # To see help\nuvicore-installer ./blog\n\n# Follow the post-installation instructions\n# Review your .env\n# Review all configs\n# Download this https://github.com/uvicore/framework code and review the guts since there are\n# no docs yet :)  Look in the tests/apps/app1 folder for a small test app to glean how to use\n# it. Sorry, it\'s not done yet, but working hard on it!\n```\n\n\n## Stable v1.0 Roadmap\n\nI am tracking issues and enhancements for v1.0 in the Issues Milestones (and on a million sticky notes).  Before this project can be considered ready for public use I need to complete the following:\n\n* Proof the system by finishing several Production sites, APIs and CLIs at my current company.\n* Write far more tests and attempt 100% coverage\n* Build the main website at https://uvicore.io\n* Complete elegant documentation (ever read Laravels docs?  Like that.) at https://docs.uvicore.io\n\n\n\n## License\n\nThe Uvicore framework is open-sourced software licensed under the MIT license.\n',
    'author': 'Matthew Reschke',
    'author_email': 'mail@mreschke.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uvicore/framework',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import uvicore
from uvicore.support.dumper import dump, dd
from uvicore.support.module import location, load
from uvicore.typing import List, Dict, OrderedDict
from uvicore.contracts import Provider
from uvicore.console import command_is


class Http(Provider):
    """Http Service Provider Mixin"""

    def web_routes(self, module: str, prefix: str, name_prefix: str = ''):
        # Default registration
        self.package.registers.defaults({'web_routes': True})

        # Allow only if running as HTTP or from certain CLI commands like package list/show...
        # if self.package.registers.web_routes and (
        #     uvicore.app.is_http or
        #     command_is('http') or
        #     command_is('package')
        # ) == False: return

        self.package.web.routes_module = module
        self.package.web.prefix = prefix
        self.package.web.name_prefix = name_prefix

        # # Import main web routes module
        # routes = load(module).object(self.package)

        # # Get name prefix from package name plus custom name prefix
        # if name_prefix:
        #     if name_prefix[0] == '.': name_prefix = name_prefix[1:]
        #     name_prefix = self.package.name + '.' + name_prefix
        # else:
        #     name_prefix = self.package.name

        # # Import the web router and create a new instance
        # from uvicore.http.routing.web_router import WebRouter  # isort:skip
        # router = WebRouter(self.package, prefix, name_prefix)

        # # Get returned router with all defined routes
        # router = routes.register(router)

        # # Merge routes into package definition
        # self.package.web.routes.merge(router.routes)

    def api_routes(self, module: str, prefix: str, name_prefix: str = 'api'):
        # Default registration
        self.package.registers.defaults({'api_routes': True})

        # Allow only if running as HTTP or from certain CLI commands like package list/show...
        # if self.package.registers.web_routes and (
        #     uvicore.app.is_http or
        #     command_is('http') or
        #     command_is('package')
        # ) == False: return

        self.package.api.routes_module = module
        self.package.api.prefix = prefix
        self.package.api.name_prefix = name_prefix

        # # Import main web routes module
        # routes = load(module).object(self.package)

        # # Get name prefix from package name plus custom name prefix
        # if name_prefix:
        #     if name_prefix[0] == '.': name_prefix = name_prefix[1:]
        #     name_prefix = self.package.name + '.' + name_prefix
        # else:
        #     name_prefix = self.package.name

        # # Import the web router and create a new instance
        # from uvicore.http.routing.api_router import ApiRouter  # isort:skip
        # router = ApiRouter(self.package, prefix, name_prefix)

        # # Get returned router with all defined routes
        # router = routes.register(router)

        # # Merge routes into package definition
        # self.package.api.routes.merge(router.routes)

    def middlewareNO(self, items: OrderedDict):
        # Don't think this should be here.  middleware is app global
        # not per package.  I add apps middleware in http service.py
        # Default registration
        self.package.registers.defaults({'middleware': True})

        # Register middleware only if allowed
        #if self.package.registers.middleware:
            #self._add_http_definition('middleware', items)

    def views(self, items: List):
        # Default registration
        self.package.registers.defaults({'views': True})

        # Register views only if allowed
        if self.package.registers.views:
            self.package.web.view_paths = items

    def assets(self, items: List):
        # Default registration
        self.package.registers.defaults({'assets': True})

        # Register assets only if allowed
        if self.package.registers.assets:
            self.package.web.asset_paths = items

    def public(self, items: List):
        self.package.web.public_paths = items


    def template(self, items: Dict):
        # Default registration - template obeys view registration
        self.package.registers.defaults({'views': True})

        # Register templates only if [views] allowed
        if self.package.registers.views:
            self.package.web.template_options = Dict(items)



from abc import ABC, abstractmethod
from uvicore.contracts.user import User
from uvicore.typing import Optional, Tuple, Dict, List, Union
try:
    from starlette.requests import HTTPConnection
except:
    HTTPConnection = None


class Authenticator(ABC):

    @abstractmethod
    async def authenticate(self, conn: HTTPConnection) -> Union[User, bool]:
        pass

    @abstractmethod
    async def retrieve_user(self, username: str, password: str, provider: Dict) -> Optional[User]:
        pass

    @abstractmethod
    def auth_header(self, request) -> Tuple[str, str, str]:
        pass



import uvicore
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd


@uvicore.provider()
class Logging(ServiceProvider):

    def register(self) -> None:
        """Register package into uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs and IoC bindings.  Configs are deep merged only after
        all packages are registered.  No real work should be performed here as it
        is very early in the bootstraping process and most internal processes are not
        instantiated yet.
        """
        # Note about logger config
        # We cannot use the standard package config/logger.py here with proper
        # app config overrides because the logger is SUPER early in the bootstrapping
        # process.  I want the log available almost first thing, even in your packages
        # service provider register() and boot() methods.  This means if the logger config
        # were in this packages config/logger.py file you wouldn't be able to override it
        # as usual from your own app.  So instead, the logger config is in your actual app config

        # Register IoC bindings
        # self.bind('Logger', 'uvicore.logging.logger._Logger',
        #     aliases=['Log', 'log', 'logger'],
        #     kwargs={'config': self.app_config.get('logger')},
        #     singleton=True
        # )

        # Set uvicore.log global
        uvicore.log = uvicore.ioc.make('uvicore.logging.logger.Logger')

    def boot(self) -> None:
        """Bootstrap package into uvicore framework.
        Boot takes place after all packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all configs.
        This is where you load views, assets, routes, commands...
        """
        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs
        # ?? course I don't have any registrations for now, maybe later
